import keras
from keras import utils
import tensorflow as tf

@utils.register_keras_serializable()
class ContextNorm(keras.layers.Layer):
    def __init__(self,
                 num_contexts,
                 priors=None,
                 axis=-1,
                 momentum=0.99,
                 epsilon=0.001,
                 center=True,
                 scale=True,
                 beta_initializer="zeros",
                 gamma_initializer="ones",
                 moving_mean_initializer="zeros",
                 moving_variance_initializer="ones",
                 beta_regularizer=None,
                 gamma_regularizer=None,
                 beta_constraint=None,
                 gamma_constraint=None,
                 synchronized=False,
                 **kwargs):
        super(ContextNorm, self).__init__(**kwargs)
        self.num_contexts = num_contexts
        self.axis = axis
        self.momentum = momentum
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = beta_initializer
        self.gamma_initializer = gamma_initializer
        self.moving_mean_initializer = moving_mean_initializer
        self.moving_variance_initializer = moving_variance_initializer
        self.beta_regularizer = beta_regularizer
        self.gamma_regularizer = gamma_regularizer
        self.beta_constraint = beta_constraint
        self.gamma_constraint = gamma_constraint
        self.synchronized = synchronized
        self.priors = priors

    def build(self, input_shape):
        self.context_norm_layers = []
        for i in range(self.num_contexts):
            self.context_norm_layers.append(keras.layers.BatchNormalization(
                axis=self.axis,
                momentum=self.momentum,
                epsilon=self.epsilon,
                center=self.center,
                scale=self.scale,
                beta_initializer=self.beta_initializer,
                gamma_initializer=self.gamma_initializer,
                moving_mean_initializer=self.moving_mean_initializer,
                moving_variance_initializer=self.moving_variance_initializer,
                beta_regularizer=self.beta_regularizer,
                gamma_regularizer=self.gamma_regularizer,
                beta_constraint=self.beta_constraint,
                gamma_constraint=self.gamma_constraint,
                trainable=True,
                name=f"context_norm_layer_{i}"
            ))
        if self.priors is None:
            self.priors = [1. / self.num_contexts] * self.num_contexts
        else:
            assert len(self.priors) == self.num_contexts, "Number of priors must match the number of groups."
        self.priors = tf.convert_to_tensor(self.priors, dtype=tf.float32)
        super(ContextNorm, self).build(input_shape)

    def call(self, inputs, training=False):
        samples, contexts = inputs
        contexts = contexts[:, 0]

        # During training, apply prior on normalized data
        if training:
            for i in range(self.num_contexts):
                indices = tf.where(tf.equal(contexts, i))
                group_data = tf.gather(samples, indices[:, 0])

                def update_samples():
                    normalized_group_data = self.context_norm_layers[i](group_data, training=True)
                    normalized_group_data *= (1. / tf.sqrt(self.priors[i]))
                    return tf.tensor_scatter_nd_update(samples, indices, normalized_group_data)

                samples = tf.cond(
                    tf.shape(indices)[0] > 0,  # Check if there are elements in this context
                    update_samples,
                    lambda: samples  # No update if no elements match the context
                )

            return samples

        # Inference
        else:
            for i in range(self.num_contexts):
                indices = tf.where(tf.equal(contexts, i))
                group_data = tf.gather(samples, indices[:, 0])

                def update_samples_inference():
                    normalized_group_data = (1. / tf.sqrt(self.priors[i])) * self.context_norm_layers[i](group_data, training=False)
                    return tf.tensor_scatter_nd_update(samples, indices, normalized_group_data)

                samples = tf.cond(
                    tf.shape(indices)[0] > 0,
                    update_samples_inference,
                    lambda: samples
                )

            return samples

    def get_config(self):
        config = {
            'num_contexts': self.num_contexts,
            'axis': self.axis,
            'momentum': self.momentum,
            'epsilon': self.epsilon,
            'center': self.center,
            'scale': self.scale,
            'beta_initializer': self.beta_initializer,
            'gamma_initializer': self.gamma_initializer,
            'moving_mean_initializer': self.moving_mean_initializer,
            'moving_variance_initializer': self.moving_variance_initializer,
            'beta_regularizer': self.beta_regularizer,
            'gamma_regularizer': self.gamma_regularizer,
            'beta_constraint': self.beta_constraint,
            'gamma_constraint': self.gamma_constraint,
            'synchronized': self.synchronized,
        }
        base_config = super(ContextNorm, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))




import tensorflow as tf
from keras.layers import Layer
import numpy as np

@utils.register_keras_serializable()
class AdaptiveContextNorm(Layer):
    def __init__(self, num_contexts, epsilon=1e-3, momentum=0.9, **kwargs):
        self.num_contexts = num_contexts
        self.epsilon = epsilon
        self.momentum = momentum
        super(AdaptiveContextNorm, self).__init__(**kwargs)

    def build(self, input_shape):
        self.input_dim = input_shape[-1]
        self.mean = self.add_weight(
            name='mean',
            shape=(self.num_contexts, self.input_dim),
            initializer=self.custom_mean_initializer,
            trainable=True
        )
        self.variance = self.add_weight(
            name='variance',
            shape=(self.num_contexts, self.input_dim),
            initializer=self.custom_variance_initializer,
            trainable=True
        )
        self.prior = self.add_weight(
            name='prior',
            shape=(self.num_contexts, 1),
            initializer=self.custom_prior_initializer,
            trainable=True
        )
        super(AdaptiveContextNorm, self).build(input_shape)

    def custom_mean_initializer(self, shape, dtype=None):
        min_value = -1.0  # Minimum value for means
        max_value = 1.0  # Maximum value for means
        return tf.random.uniform(shape=shape, minval=min_value, maxval=max_value, dtype=dtype)

    def custom_variance_initializer(self, shape, dtype=None):
        min_value = 0.001  # Minimum value for variance
        max_value = 0.01  # Maximum value for variance
        return tf.random.uniform(shape=shape, minval=min_value, maxval=max_value, dtype=dtype)

    def custom_prior_initializer(self, shape, dtype=None):
        initial_values = tf.random.uniform(shape, minval=0.01, maxval=0.99, dtype=dtype)
        normalized_values = initial_values / tf.reduce_sum(initial_values, axis=0)
        return normalized_values

    def get_config(self):
        config = super(AdaptiveContextNorm, self).get_config().copy()
        config.update(
            {
                "num_contexts": self.num_contexts,
                "epsilon": self.epsilon,
                "momentum": self.momentum
            })
        return config

    def compute_output_shape(self, input_shape):
        return input_shape

    def call(self, inputs, training=False):
        if training is None:
            training = tf.keras.backend.learning_phase()
        x = inputs
        normalized_x = tf.zeros_like(x)
        mean = self.mean
        var = self.variance
        prior = self.prior
        num_expand_dims = len(x.shape) - 2

        for _ in range(num_expand_dims):
            mean = tf.expand_dims(mean, axis=1)
            var = tf.expand_dims(var, axis=1)

        var = tf.math.softplus(var)
        prior = tf.math.softmax(prior)

        for k in range(self.num_contexts):
            mean_k = mean[k, :]
            var_k = var[k, :]
            prior_k = prior[k, :]

            p_x_given_k = prior_k * tf.math.exp(-0.5 * ((x - mean_k) * (1 / (var_k + self.epsilon)) * (x - mean_k)))

            p_x_given_i = tf.zeros_like(p_x_given_k)
            for i in range(self.num_contexts):
                mean_i = mean[i, :]
                var_i = var[i, :]

                posterior_proba = prior[i] * tf.math.exp(
                    -0.5 * ((x - mean_i) * (1 / (var_i + self.epsilon)) * (x - mean_i)))

                p_x_given_i += posterior_proba

            tau_k = p_x_given_k / (p_x_given_i + self.epsilon)

            if training:
                sum_tau_k = tf.reduce_sum(tau_k, axis=list(range(x.ndim))[:-1])
                hat_tau_k = tau_k / (sum_tau_k + self.epsilon)

                prod = hat_tau_k * x
                expectation = tf.reduce_mean(prod, axis=list(range(x.ndim))[1:-1])
                shape1 = x.shape
                shape2 = expectation.shape
                num_extra_dims = len(shape1) - len(shape2)
                for _ in range(num_extra_dims):
                    expectation = tf.expand_dims(expectation, axis=1)
                v_i_k = x - expectation

                prod_bis = hat_tau_k * (prod * prod)
                variance = tf.reduce_mean(prod_bis, axis=list(range(x.ndim))[1:-1])
                shape1 = v_i_k.shape
                shape2 = variance.shape
                num_extra_dims = len(shape1) - len(shape2)
                for _ in range(num_extra_dims):
                    variance = tf.expand_dims(variance, axis=1)
                hat_x_i_k = v_i_k / tf.sqrt(variance + self.epsilon)

                hat_x_i = (tau_k / tf.math.sqrt(prior_k + self.epsilon)) * hat_x_i_k

                normalized_x += hat_x_i

                updated_mean = tf.reduce_mean(hat_x_i, axis=list(range(x.ndim))[:-1])
                updated_var = tf.math.reduce_variance(hat_x_i, axis=list(range(x.ndim))[:-1])
                updated_prior = tf.reduce_mean(tau_k)

                self.mean[k, :].assign(self.momentum * self.mean[k, :] + (1.0 - self.momentum) * updated_mean)
                self.variance[k, :].assign(self.momentum * self.variance[k, :] + (1.0 - self.momentum) * updated_var)
                self.prior[k, 0].assign(self.momentum * self.prior[k, 0] + (1.0 - self.momentum) * updated_prior)

            else:
                hat_x_i_k = (x - mean_k) / (tf.math.sqrt(var_k + self.epsilon))
                hat_x_i = (tau_k / tf.math.sqrt(prior_k + self.epsilon)) * hat_x_i_k
                normalized_x += hat_x_i

        return normalized_x




import keras
from keras import layers, utils
import tensorflow as tf

@utils.register_keras_serializable()
class ContextExtendedNorm(layers.Layer):
    def __init__(self, num_contexts, epsilon=1e-3, **kwargs):
        """
        Initialize the Context Extended Normalization layer.

        Parameters:
        :param num_contexts: The number of contexts (prior knowledge)
        :param epsilon: A small positive value to prevent division by zero during normalization.
        """
        self.num_contexts = num_contexts
        self.epsilon = epsilon
        super(ContextExtendedNorm, self).__init__(**kwargs)

    def build(self, input_shape):
        """
        Build the layer by creating sub-layers for learning initial mean and standard deviation.

        :param input_shape: The shape of the layer's input.

        This method initializes the layers for learning initial mean and standard deviation, based on the input shape.
        """
        self.input_dim = input_shape[0][-1]

        # Create weights for initial mean and standard deviation
        self.initial_mean = self.add_weight(
            name='initial_mean',
            shape=(self.num_contexts, self.input_dim),
            initializer='glorot_uniform',
            trainable=True
        )
        self.initial_std = self.add_weight(
            name='initial_std',
            shape=(self.num_contexts, self.input_dim),
            initializer='glorot_uniform',
            trainable=True
        )
        super(ContextExtendedNorm, self).build(input_shape)
        
    def get_config(self):
        config = super().get_config().copy()
        config.update({
            "num_contexts": self.num_contexts,
            "epsilon": self.epsilon
        })
        return config

    def call(self, inputs):
        """
        Apply the Context Extended Normalization to the input data.

        :param inputs: A tuple of (x, context_id) where x is the data to be normalized, and context_id is the context identifier. Context identifier must be int32 format.

        :return normalized_x: The normalized output data.
        """
        x, context_id = inputs

        # Extract context indices from context_id
        indices = context_id[:, 0]

        # Gather initial mean and standard deviation based on context indices
        mean = tf.gather(self.initial_mean, indices)
        std = tf.gather(self.initial_std, indices)

        # Ensure standard deviation is positive
        std = tf.exp(std)

        # Determine the number of dimensions to expand
        num_expand_dims = len(x.shape) - 2

        # Expand mean and std dimensions accordingly
        for _ in range(num_expand_dims):
            mean = tf.expand_dims(mean, axis=1)
            std = tf.expand_dims(std, axis=1)

        # Perform normalization
        normalized_x = (x - mean) / (std + self.epsilon)

        return normalized_x




from .ContextNorm import ContextNorm
from .ContextExtendedNorm import ContextExtendedNorm
from .AdaptiveContextNorm import AdaptiveContextNorm

__all__ = [
    "ContextNorm",
    "ContextExtendedNorm",
    "AdaptiveContextNorm",
]



from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="keras_context_based_norm",
    version="1.0.0",
    author="Bilal FAYE",
    author_email="faye@lipn.univ-paris13.fr",
    url="https://github.com/b-faye/prior-based-norm",
    description="Prior-Based Normalization provides versatile normalization layers for deep neural networks, including ContextNorm, ContextExtendedNorm and AdaptiveContextNorm versions. Enhance model generalization and robustness by efficiently integrating prior knowledge.",    
    packages=find_packages(),
    readme="README.md",
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=["tensorflow >= 2.17.0"],
    python_requires=">=3.10",
    classifiers=[
    "Programming Language :: Python :: 3.10",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Software Development :: Libraries :: Python Modules"
    ]


)
